/*
 * Copyright (c) 2015      Intel, Inc.  All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef PMIX_MUNGE_H
#define PMIX_MUNGE_H

BEGIN_C_DECLS

#include <src/include/pmix_config.h>


extern pmix_sec_base_module_t pmix_munge_module;

END_C_DECLS

#endif
