/*
 * Copyright (c) 2011-2013 Los Alamos National Security, LLC.  All rights
 *                         reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include "orte_config.h"

#include "rml_oob.h"

int
orte_rml_oob_ping(const char* uri,
                  const struct timeval* tv)
{
    return ORTE_ERR_NOT_SUPPORTED;
}
