/*
 * Copyright (c) 2012      Los Alamos National Security, LLC.
 *                         All rights reserved
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef ORTE_RMAPS_STAGED_H
#define ORTE_RMAPS_STAGED_H

#include "orte_config.h"

#include "orte/mca/rmaps/rmaps.h"

BEGIN_C_DECLS

ORTE_MODULE_DECLSPEC extern orte_rmaps_base_component_t mca_rmaps_staged_component;
extern orte_rmaps_base_module_t orte_rmaps_staged_module;

END_C_DECLS

#endif
