/*
 * Copyright (c) 2014      Intel, Inc. All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef OPAL_SEC_BASIC_H
#define OPAL_SEC_BASIC_H

#include "opal/mca/sec/sec.h"

BEGIN_C_DECLS


OPAL_MODULE_DECLSPEC extern opal_sec_base_component_t mca_sec_basic_component;
OPAL_DECLSPEC extern opal_sec_base_module_t opal_sec_basic_module;

END_C_DECLS

#endif /* OPAL_SEC_BASIC_H */
