/*
 * Copyright (c) 2016      Mellanox Technologies, Inc.
 *                         All rights reserved.
 * Copyright (c) 2016      IBM Corporation.  All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef PMIX_INCLUDE_CONFIG_H
#define PMIX_INCLUDE_CONFIG_H

#include <pmix/autogen/config.h>

#endif /* PMIX_INCLUDE_CONFIG_H */
