! -*- f90 -*-
!
! Copyright (c) 2010-2012 Cisco Systems, Inc.  All rights reserved.
! Copyright (c) 2009-2012 Los Alamos National Security, LLC.
!                         All rights reserved.
! $COPYRIGHT$

subroutine MPI_Win_get_name_f08(win,win_name,resultlen,ierror)
   use, intrinsic :: ISO_C_BINDING, only : C_CHAR
   use :: mpi_f08_types, only : MPI_Win, MPI_MAX_OBJECT_NAME
   use :: mpi_f08, only : ompi_win_get_name_f
   implicit none
   TYPE(MPI_Win), INTENT(IN) :: win
   CHARACTER(KIND=C_CHAR), DIMENSION(*), INTENT(OUT) :: win_name
   INTEGER, INTENT(OUT) :: resultlen
   INTEGER, OPTIONAL, INTENT(OUT) :: ierror
   integer :: c_ierror

   call ompi_win_get_name_f(win%MPI_VAL,win_name,resultlen,c_ierror,len(win_name))
   if (present(ierror)) ierror = c_ierror

end subroutine MPI_Win_get_name_f08
